from pydantic import BaseModel, Field, AnyUrl
import yaml
from typing import Literal, Optional
from pathlib import Path
from metagen.helpers import make_hash
from metagen.base import singleton

BASE_CONFIG_FILE = Path(__file__).parent / 'config.yaml'


class HashableBaseModel(BaseModel):
    """Helper class implementing hash into the base model"""

    def __hash__(self) -> int:
        return make_hash({k: v for k, v in self.__dict__.items()})


class RegisterConfig(HashableBaseModel):
    registerName: Literal['pandas', 'dict']

    def __str__(self):
        items = [f"\t- {k} : {v}\n" for k, v in self.__dict__.items()]
        return ''.join(items)


class ImporterConfig(HashableBaseModel):
    path: Optional[str]
    instance_url: Optional[str]
    host: Optional[str]

    def __str__(self):
        items = [f"\t- {k} : {v}\n" for k, v in self.__dict__.items()]
        return ''.join(items)


@singleton
class Config(HashableBaseModel):
    register_setting: Optional[RegisterConfig] = Field(default_factory=RegisterConfig)
    importer_setting: Optional[ImporterConfig] = Field(default_factory=ImporterConfig)

    def __init__(self):
        super().__init__(**load_yaml(BASE_CONFIG_FILE))

    def __setattr__(self, name, value):
        """Called when attribute is set. Automatically save the setting"""
        setattr(self, name, value)
        dump_yaml(Path(BASE_CONFIG_FILE), self.dict())

    def __str__(self):
        items = [f"{k} : {str(v)}" for k, v in self.__dict__.items()]
        return ''.join([f"{k} : \n {str(v)}" for k, v in self.__dict__.items()])


def load_yaml(path: str) -> dict:
    with open(path, 'r') as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def dump_yaml(path: Path, data: dict) -> None:
    with open(path, 'w') as file:
        file.write(yaml.dump(data, Dumper=yaml.Dumper))






from metagen.main import PTRMetagen
from metagen.metadata import Application, Configuration, Scope, Place, Period, LayerTemplate, Attribute, SpatialWMS, \
    RelationSpatial, SpatialAttribute, SpatialWMTS, SpatialVector, Style, Case, RelationAttribute, Tag
from metagen.config.config import Config

__version__ = '1.1.17'


from abc import ABC, abstractmethod
from uuid import UUID
from typing import Type, Any, List, Dict
from pydantic import BaseModel, Field

from metagen.base import LeafABC, BaseModelArbitrary
from metagen.helpers import Singleton


class RegisterABC(BaseModelArbitrary, ABC):

    @abstractmethod
    def get_elements(self) -> List[Type[LeafABC]]:
        pass

    @abstractmethod
    def add(self, element: Type[LeafABC]) -> None:
        pass

    @abstractmethod
    def check_register(self, element: Type[LeafABC]) -> bool:
        pass

    @abstractmethod
    def get_by_uuid(self, uuid: UUID) -> Type[LeafABC]:
        pass


# register
class DictRegister(RegisterABC, Singleton):
    uuid: dict = Field(default_factory=dict)

    @property
    def hashes(self):
        return {hash(element): element for element in self.uuid.values()}

    def get_elements(self) -> List[Type[LeafABC]]:
        return [element for element in self.uuid.values()]

    def add(self, element: Type[LeafABC]) -> None:
        # TODO resolve hash and uuid check
        if not self.check_register(element):
            if element.key not in self.uuid:
                self.uuid.update({str(element.key): element})
            else:
                raise KeyError(f'UUID duplicity conflict: element {element} with UUID {element.key} exist in register')
        else:
            raise ValueError(f'Hash duplicity conflict: "{element.__class__.__name__}" with '
                             f'key: {element.key} and hash: {hash(element)} already exist')

    def check_register(self, element: Type[LeafABC]) -> bool:
        return hash(element) in self.hashes

    def get_by_uuid(self, uuid: str) -> Type[LeafABC]:
        if UUID(uuid):
            return self.uuid.get(uuid)

    def get_by_hash(self, hash: int) -> Type[LeafABC]:
        return self.hashes.get(hash)


class RegisterFactory(BaseModel):
    registers: Dict[str, Type[RegisterABC]] = Field(default_factory=dict)

    def add(self, registerName: str, registerType: Type[RegisterABC]) -> None:
        self.registers.update({registerName: registerType})

    def get(self, registerName: str, **ignore) -> Type[RegisterABC]:
        return self.registers[registerName]


register_factory = RegisterFactory()
register_factory.add(registerName='dict', registerType=DictRegister)
